from dataclasses import dataclass, field


@dataclass
class EntityResolutionResult:
    """Ground truth is expressed in terms of one or more mathematical models."""

    """Entity resolution result expressed in terms of the Algebraic model.

    The algebraic model for entity resolution views entity resolution as an equivalence algebraic
    relation between elements of the same data source with the properties of reflexivity, symmetry
    and distributivity. Such a relation partitions the original data source in a unique way.
    Therefore, in this model, the entity resolution over a data source is the partitioning of
    equivalent elements over the same data source. This model was proposed by John R. Talburt in
    the book
    `Entity Resolution and Information Quality<https://books.google.ro/books?id=tIB0IZYR8V8C&dq=john+r.+talburt>`_.
    """
    algebraic: list[list[tuple]] = field(default_factory=list, init=False)
    """Entity resolution result in terms of the Fellegi-Sunter probabilistic model.

    The Fellegi-Sunter model is the oldest mathematical model for entity resolution. Its origins
    can be traced back to the paper written by Ivan Fellegi and Alan Sunter:
    `A Theory for Record Linkage<https://www.tandfonline.com/doi/abs/10.1080/01621459.1969.10501049>`_.
    """
    fsm: list[tuple[tuple]] = field(default_factory=list, init=False)
    """Entity resolution result expressed in terms of the SERF functional model.

    The SERF (Standard Entity Resolution Framework) mathematical model proposes a way of looking
    at entity resolution as if it were the result of applying two types of functions (match
    functions and merge functions) over two data sources with unique records (a data set). The model
    builds new information records from pairs of existing information records by using a "domination"
    relationship. The "dominating" information records take the place of the "dominated" information
    records until there aren't any more records in the two data sets that can create new "dominating"
    records. The information records obtained by this process make up a data set that is said to be
    the entity resolution over the two initial data sources. The SERF ground truth depends on which
    ``merge`` function is used.

    It was proposed by Omar Benjelloun et al. in the paper
    `Swoosh: a generic approach to entity resolution<https://link.springer.com/article/10.1007/s00778-008-0098-x>`_.
    """
    serf: list[tuple] = field(default_factory=list, init=False)



from matchescu.typing._data import Record, RecordSampler, DataSource, Trait, TRecord
from matchescu.typing._references import (
    EntityReferenceIdentifier,
    EntityReference,
    EntityReferenceIdFactory,
    EntityProfile,
    RecordAdapter,
)


__all__ = [
    "DataSource",
    "EntityReference",
    "EntityReferenceIdentifier",
    "EntityReferenceIdFactory",
    "EntityProfile",
    "Record",
    "RecordSampler",
    "RecordAdapter",
    "Trait",
    "TRecord",
]



from collections.abc import Iterable
from typing import Any, Iterator


class Record:
    def __init__(self, value: Iterable) -> None:
        if isinstance(value, Record):
            self.__values = value.__values
            self.__attr_names = value.__attr_names
        else:
            tuples = list(self.__init_data(value))
            self.__values = tuple(x[1] for x in tuples)
            self.__attr_names = {x[0]: i for i, x in enumerate(tuples)}

    @staticmethod
    def merge(records: Iterable["Record"]) -> "Record":
        merge_record = {}
        for record in records:
            merge_record.update(
                {k: record.__values[i] for k, i in record.__attr_names.items()}
            )
        return Record(merge_record)

    @staticmethod
    def __get_attr_key(key: str | int) -> str:
        return f"column_{key}" if isinstance(key, int) else key

    def __init_data(self, value: Iterable) -> Iterable[tuple]:
        if isinstance(value, dict):
            return ((self.__get_attr_key(k), v) for k, v in value.items())
        elif isinstance(value, (tuple, list, set)):
            return ((self.__get_attr_key(i), v) for i, v in enumerate(value, start=1))
        raise ValueError(
            f"can't initialize data record from '{type(value).__name__}' values"
        )

    def __getitem__(self, key: str | int) -> Any:
        if isinstance(key, str):
            return (
                self.__values[self.__attr_names[key]]
                if key in self.__attr_names
                else None
            )
        elif isinstance(key, int):
            return self.__values[key]
        raise ValueError(f"can't get data record using '{type(key).__name__}' keys")

    def __getattr__(self, key: str) -> Any:
        if key not in self.__attr_names:
            raise AttributeError(f"Record attribute '{key}' not found")
        return self[key]

    def __len__(self) -> int:
        return len(self.__values)

    def __iter__(self) -> Iterator[Any]:
        return iter(self.__values)



from typing import TypeVar, Hashable, Generic


T = TypeVar("T", bound=Hashable)


class EquivalenceClassPartitioner(Generic[T]):
    def __init__(self, items: list[T]):
        self._rank = {item: 0 for item in items}
        self._parent = {item: item for item in items}
        self._items = items

    def _find(self, x: T) -> T:
        if self._parent[x] == x:
            return x
        # path compression
        self._parent[x] = self._find(self._parent[x])
        return self._parent[x]

    def _union(self, x: T, y: T) -> None:
        x_root = self._find(x)
        y_root = self._find(y)

        if x_root == y_root:
            return

        if self._rank[x_root] < self._rank[y_root]:
            self._parent[x_root] = y_root
        elif self._rank[y_root] < self._rank[x_root]:
            self._parent[y_root] = x_root
        else:
            # does not matter which goes where
            # make sure we increase the correct rank
            self._parent[y_root] = x_root
            self._rank[x_root] += 1

    def __call__(self, pairs: list[tuple[T, T]]) -> list[list[T]]:
        for x, y in pairs:
            self._union(x, y)
        classes = {item: dict() for item in self._items}
        for item in self._items:
            classes[self._find(item)][item] = None
        return [list(eq_class) for eq_class in classes.values() if len(eq_class) > 0]


def compute_partition(
    all_references: list, equivalence: list[tuple]
) -> list[list[tuple]]:
    """Partition an input set according to an equivalence relation.

    The algorithm uses the Union-Find or DFU data structure.

    :param all_references: all entity references that make up the input domain
    :param equivalence: pairs that define which items are equivalent to other items
    :return: a list of lists of tuples, where:

    * each tuple is an entity reference
    * tuples are grouped in clusters which are represented as lists
    * the final result is a list of clusters
    """
    partitioner = EquivalenceClassPartitioner(all_references)
    result = partitioner(equivalence)
    return result
